@@ -1,20 +1,61 @@
 #include "cpr/session.h"
 
-#include <algorithm>
+#include <atomic>
+#include <cassert>
+#include <cstdint>
 #include <cstdlib>
 #include <cstring>
 #include <fstream>
-#include <functional>
 #include <iostream>
+#include <memory>
+#include <optional>
 #include <stdexcept>
 #include <string>
+#include <utility>
+#include <variant>
+#include <vector>
 
 #include <curl/curl.h>
+#include <curl/curlver.h>
+#include <curl/easy.h>
+#include <curl/system.h>
 
+#include "cpr/accept_encoding.h"
 #include "cpr/async.h"
+#include "cpr/auth.h"
+#include "cpr/bearer.h"
+#include "cpr/body.h"
+#include "cpr/callback.h"
+#include "cpr/connect_timeout.h"
+#include "cpr/cookies.h"
 #include "cpr/cprtypes.h"
+#include "cpr/curlholder.h"
+#include "cpr/error.h"
+#include "cpr/file.h"
+#include "cpr/filesystem.h" // IWYU pragma: keep
+#include "cpr/http_version.h"
 #include "cpr/interceptor.h"
+#include "cpr/interface.h"
+#include "cpr/limit_rate.h"
+#include "cpr/local_port.h"
+#include "cpr/local_port_range.h"
+#include "cpr/low_speed.h"
+#include "cpr/multipart.h"
+#include "cpr/parameters.h"
+#include "cpr/payload.h"
+#include "cpr/proxies.h"
+#include "cpr/proxyauth.h"
+#include "cpr/range.h"
+#include "cpr/redirect.h"
+#include "cpr/reserve_size.h"
+#include "cpr/resolve.h"
+#include "cpr/response.h"
+#include "cpr/ssl_options.h"
+#include "cpr/timeout.h"
+#include "cpr/unix_socket.h"
+#include "cpr/user_agent.h"
 #include "cpr/util.h"
+#include "cpr/verbose.h"
 
 #if SUPPORT_CURLOPT_SSL_CTX_FUNCTION
 #include "cpr/ssl_ctx.h"
@@ -31,13 +72,13 @@ constexpr long OFF = 0L;
 
 CURLcode Session::DoEasyPerform() {
     if (isUsedInMultiPerform) {
-        std::cerr << "curl_easy_perform cannot be executed if the CURL handle is used in a MultiPerform." << std::endl;
+        std::cerr << "curl_easy_perform cannot be executed if the CURL handle is used in a MultiPerform.\n";
         return CURLcode::CURLE_FAILED_INIT;
     }
     return curl_easy_perform(curl_->handle);
 }
 
-void Session::SetHeaderInternal() {
+void Session::prepareHeader() {
     curl_slist* chunk = nullptr;
     for (const std::pair<const std::string, std::string>& item : header_) {
         std::string header_string = item.first;
@@ -97,14 +138,17 @@ Session::Session() : curl_(new CurlHolder()) {
     curl_easy_setopt(curl_->handle, CURLOPT_NOSIGNAL, 1L);
 #endif
 
-#if LIBCURL_VERSION_NUM >= 0x071900
+#if LIBCURL_VERSION_NUM >= 0x071900 // 7.25.0
     curl_easy_setopt(curl_->handle, CURLOPT_TCP_KEEPALIVE, 1L);
 #endif
+    current_interceptor_ = interceptors_.end();
+    first_interceptor_ = interceptors_.end();
 }
 
 Response Session::makeDownloadRequest() {
-    if (!interceptors_.empty()) {
-        return intercept();
+    const std::optional<Response> r = intercept();
+    if (r.has_value()) {
+        return r.value();
     }
 
     const CURLcode curl_error = DoEasyPerform();
@@ -112,12 +156,13 @@ Response Session::makeDownloadRequest() {
     return CompleteDownload(curl_error);
 }
 
-void Session::prepareCommon() {
+void Session::prepareCommonShared() {
     assert(curl_->handle);
 
     // Set Header:
-    SetHeaderInternal();
+    prepareHeader();
 
+    // URL parameter:
     const std::string parametersContent = parameters_.GetContent(*curl_);
     if (!parametersContent.empty()) {
         const Url new_url{url_ + "?" + parametersContent};
@@ -131,12 +176,12 @@ void Session::prepareCommon() {
     if (proxies_.has(protocol)) {
         curl_easy_setopt(curl_->handle, CURLOPT_PROXY, proxies_[protocol].c_str());
         if (proxyAuth_.has(protocol)) {
-            curl_easy_setopt(curl_->handle, CURLOPT_PROXYAUTH, CURLAUTH_ANY);
-            curl_easy_setopt(curl_->handle, CURLOPT_PROXYUSERPWD, proxyAuth_[protocol]);
+            curl_easy_setopt(curl_->handle, CURLOPT_PROXYUSERNAME, proxyAuth_.GetUsername(protocol));
+            curl_easy_setopt(curl_->handle, CURLOPT_PROXYPASSWORD, proxyAuth_.GetPassword(protocol));
         }
     }
 
-#if LIBCURL_VERSION_NUM >= 0x072100
+#if LIBCURL_VERSION_NUM >= 0x071506 // 7.21.6
     if (acceptEncoding_.empty()) {
         // Enable all supported built-in compressions
         curl_easy_setopt(curl_->handle, CURLOPT_ACCEPT_ENCODING, "");
@@ -148,7 +193,7 @@ void Session::prepareCommon() {
     }
 #endif
 
-#if LIBCURL_VERSION_NUM >= 0x077100
+#if LIBCURL_VERSION_NUM >= 0x071900 // 7.25.0
 #if SUPPORT_SSL_NO_REVOKE
     // NOLINTNEXTLINE (google-runtime-int)
     long bitmask{0};
@@ -156,8 +201,10 @@ void Session::prepareCommon() {
     const bool noRevoke = bitmask & CURLSSLOPT_NO_REVOKE;
 #endif
 
+#if LIBCURL_VERSION_NUM >= 0x074700 // 7.71.0
     // Fix loading certs from Windows cert store when using OpenSSL:
     curl_easy_setopt(curl_->handle, CURLOPT_SSL_OPTIONS, CURLSSLOPT_NATIVE_CA);
+#endif
 
 // Ensure SSL no revoke is still set
 #if SUPPORT_SSL_NO_REVOKE
@@ -169,53 +216,50 @@ void Session::prepareCommon() {
 
     curl_->error[0] = '\0';
 
+    // Clear the response
     response_string_.clear();
     if (response_string_reserve_size_ > 0) {
         response_string_.reserve(response_string_reserve_size_);
     }
-    header_string_.clear();
-    if (!this->writecb_.callback) {
+
+    // Enable so we are able to retrieve certificate information:
+    curl_easy_setopt(curl_->handle, CURLOPT_CERTINFO, 1L);
+}
+
+void Session::prepareCommon() {
+    assert(curl_->handle);
+
+    // Everything else:
+    prepareCommonShared();
+
+    // Set Content:
+    prepareBodyPayloadOrMultipart();
+
+    if (!cbs_->writecb_.callback) {
         curl_easy_setopt(curl_->handle, CURLOPT_WRITEFUNCTION, cpr::util::writeFunction);
         curl_easy_setopt(curl_->handle, CURLOPT_WRITEDATA, &response_string_);
     }
-    if (!this->headercb_.callback) {
+
+    header_string_.clear();
+    if (!cbs_->headercb_.callback) {
         curl_easy_setopt(curl_->handle, CURLOPT_HEADERFUNCTION, cpr::util::writeFunction);
         curl_easy_setopt(curl_->handle, CURLOPT_HEADERDATA, &header_string_);
     }
-
-    // Enable so we are able to retrive certificate information:
-    curl_easy_setopt(curl_->handle, CURLOPT_CERTINFO, 1L);
 }
 
 void Session::prepareCommonDownload() {
     assert(curl_->handle);
 
+    // Everything else:
+    prepareCommonShared();
+
     // Set Header:
-    SetHeaderInternal();
-
-    const std::string parametersContent = parameters_.GetContent(*curl_);
-    if (!parametersContent.empty()) {
-        const Url new_url{url_ + "?" + parametersContent};
-        curl_easy_setopt(curl_->handle, CURLOPT_URL, new_url.c_str());
-    } else {
-        curl_easy_setopt(curl_->handle, CURLOPT_URL, url_.c_str());
-    }
-
-    const std::string protocol = url_.str().substr(0, url_.str().find(':'));
-    if (proxies_.has(protocol)) {
-        curl_easy_setopt(curl_->handle, CURLOPT_PROXY, proxies_[protocol].c_str());
-        if (proxyAuth_.has(protocol)) {
-            curl_easy_setopt(curl_->handle, CURLOPT_PROXYAUTH, CURLAUTH_ANY);
-            curl_easy_setopt(curl_->handle, CURLOPT_PROXYUSERPWD, proxyAuth_[protocol]);
-        }
-    }
-
-    curl_->error[0] = '\0';
+    prepareHeader();
 
     header_string_.clear();
-    if (headercb_.callback) {
+    if (cbs_->headercb_.callback) {
         curl_easy_setopt(curl_->handle, CURLOPT_HEADERFUNCTION, cpr::util::headerUserFunction);
-        curl_easy_setopt(curl_->handle, CURLOPT_HEADERDATA, &headercb_);
+        curl_easy_setopt(curl_->handle, CURLOPT_HEADERDATA, &cbs_->headercb_);
     } else {
         curl_easy_setopt(curl_->handle, CURLOPT_HEADERFUNCTION, cpr::util::writeFunction);
         curl_easy_setopt(curl_->handle, CURLOPT_HEADERDATA, &header_string_);
@@ -223,11 +267,13 @@ void Session::prepareCommonDownload() {
 }
 
 Response Session::makeRequest() {
-    if (!interceptors_.empty()) {
-        return intercept();
+    const std::optional<Response> r = intercept();
+    if (r.has_value()) {
+        return r.value();
     }
 
     const CURLcode curl_error = DoEasyPerform();
+
     return Complete(curl_error);
 }
 
@@ -237,46 +283,46 @@ void Session::SetLimitRate(const LimitRate& limit_rate) {
 }
 
 void Session::SetReadCallback(const ReadCallback& read) {
-    readcb_ = read;
+    cbs_->readcb_ = read;
     curl_easy_setopt(curl_->handle, CURLOPT_INFILESIZE_LARGE, read.size);
     curl_easy_setopt(curl_->handle, CURLOPT_POSTFIELDSIZE_LARGE, read.size);
     curl_easy_setopt(curl_->handle, CURLOPT_READFUNCTION, cpr::util::readUserFunction);
-    curl_easy_setopt(curl_->handle, CURLOPT_READDATA, &readcb_);
+    curl_easy_setopt(curl_->handle, CURLOPT_READDATA, &cbs_->readcb_);
     chunkedTransferEncoding_ = read.size == -1;
 }
 
 void Session::SetHeaderCallback(const HeaderCallback& header) {
     curl_easy_setopt(curl_->handle, CURLOPT_HEADERFUNCTION, cpr::util::headerUserFunction);
-    headercb_ = header;
-    curl_easy_setopt(curl_->handle, CURLOPT_HEADERDATA, &headercb_);
+    cbs_->headercb_ = header;
+    curl_easy_setopt(curl_->handle, CURLOPT_HEADERDATA, &cbs_->headercb_);
 }
 
 void Session::SetWriteCallback(const WriteCallback& write) {
     curl_easy_setopt(curl_->handle, CURLOPT_WRITEFUNCTION, cpr::util::writeUserFunction);
-    writecb_ = write;
-    curl_easy_setopt(curl_->handle, CURLOPT_WRITEDATA, &writecb_);
+    cbs_->writecb_ = write;
+    curl_easy_setopt(curl_->handle, CURLOPT_WRITEDATA, &cbs_->writecb_);
 }
 
 void Session::SetProgressCallback(const ProgressCallback& progress) {
-    progresscb_ = progress;
+    cbs_->progresscb_ = progress;
     if (isCancellable) {
-        cancellationcb_.SetProgressCallback(progresscb_);
+        cbs_->cancellationcb_.SetProgressCallback(cbs_->progresscb_);
         return;
     }
-#if LIBCURL_VERSION_NUM < 0x072000
+#if LIBCURL_VERSION_NUM < 0x072000 // 7.32.0
     curl_easy_setopt(curl_->handle, CURLOPT_PROGRESSFUNCTION, cpr::util::progressUserFunction<ProgressCallback>);
-    curl_easy_setopt(curl_->handle, CURLOPT_PROGRESSDATA, &progresscb_);
+    curl_easy_setopt(curl_->handle, CURLOPT_PROGRESSDATA, &cbs_->progresscb_);
 #else
     curl_easy_setopt(curl_->handle, CURLOPT_XFERINFOFUNCTION, cpr::util::progressUserFunction<ProgressCallback>);
-    curl_easy_setopt(curl_->handle, CURLOPT_XFERINFODATA, &progresscb_);
+    curl_easy_setopt(curl_->handle, CURLOPT_XFERINFODATA, &cbs_->progresscb_);
 #endif
     curl_easy_setopt(curl_->handle, CURLOPT_NOPROGRESS, 0L);
 }
 
 void Session::SetDebugCallback(const DebugCallback& debug) {
     curl_easy_setopt(curl_->handle, CURLOPT_DEBUGFUNCTION, cpr::util::debugUserFunction);
-    debugcb_ = debug;
-    curl_easy_setopt(curl_->handle, CURLOPT_DEBUGDATA, &debugcb_);
+    cbs_->debugcb_ = debug;
+    curl_easy_setopt(curl_->handle, CURLOPT_DEBUGDATA, &cbs_->debugcb_);
     curl_easy_setopt(curl_->handle, CURLOPT_VERBOSE, 1L);
 }
 
@@ -340,6 +386,10 @@ void Session::SetAuth(const Authentication& auth) {
             curl_easy_setopt(curl_->handle, CURLOPT_HTTPAUTH, CURLAUTH_NTLM);
             curl_easy_setopt(curl_->handle, CURLOPT_USERPWD, auth.GetAuthString());
             break;
+        case AuthMode::NEGOTIATE:
+            curl_easy_setopt(curl_->handle, CURLOPT_HTTPAUTH, CURLAUTH_NEGOTIATE);
+            curl_easy_setopt(curl_->handle, CURLOPT_USERPWD, auth.GetAuthString());
+            break;
     }
 }
 
@@ -348,17 +398,11 @@ void Session::SetUserAgent(const UserAgent& ua) {
 }
 
 void Session::SetPayload(const Payload& payload) {
-    hasBodyOrPayload_ = true;
-    const std::string content = payload.GetContent(*curl_);
-    curl_easy_setopt(curl_->handle, CURLOPT_POSTFIELDSIZE_LARGE, static_cast<curl_off_t>(content.length()));
-    curl_easy_setopt(curl_->handle, CURLOPT_COPYPOSTFIELDS, content.c_str());
+    content_ = payload;
 }
 
 void Session::SetPayload(Payload&& payload) {
-    hasBodyOrPayload_ = true;
-    const std::string content = payload.GetContent(*curl_);
-    curl_easy_setopt(curl_->handle, CURLOPT_POSTFIELDSIZE_LARGE, static_cast<curl_off_t>(content.length()));
-    curl_easy_setopt(curl_->handle, CURLOPT_COPYPOSTFIELDS, content.c_str());
+    content_ = std::move(payload);
 }
 
 void Session::SetProxies(const Proxies& proxies) {
@@ -378,51 +422,11 @@ void Session::SetProxyAuth(const ProxyAuthentication& proxy_auth) {
 }
 
 void Session::SetMultipart(const Multipart& multipart) {
-    // Make sure, we have a empty multipart to start with:
-    if (curl_->multipart) {
-        curl_mime_free(curl_->multipart);
-    }
-    curl_->multipart = curl_mime_init(curl_->handle);
-
-    // Add all multipart pieces:
-    for (const Part& part : multipart.parts) {
-        if (part.is_file) {
-            for (const File& file : part.files) {
-                curl_mimepart* mimePart = curl_mime_addpart(curl_->multipart);
-                if (!part.content_type.empty()) {
-                    curl_mime_type(mimePart, part.content_type.c_str());
-                }
-
-                curl_mime_filedata(mimePart, file.filepath.c_str());
-                curl_mime_name(mimePart, part.name.c_str());
-
-                if (file.hasOverridenFilename()) {
-                    curl_mime_filename(mimePart, file.overriden_filename.c_str());
-                }
-            }
-        } else {
-            curl_mimepart* mimePart = curl_mime_addpart(curl_->multipart);
-            if (!part.content_type.empty()) {
-                curl_mime_type(mimePart, part.content_type.c_str());
-            }
-            if (part.is_buffer) {
-                // Do not use formdata, to prevent having to use reinterpreter_cast:
-                curl_mime_name(mimePart, part.name.c_str());
-                curl_mime_data(mimePart, part.data, part.datalen);
-                curl_mime_filename(mimePart, part.value.c_str());
-            } else {
-                curl_mime_name(mimePart, part.name.c_str());
-                curl_mime_data(mimePart, part.value.c_str(), CURL_ZERO_TERMINATED);
-            }
-        }
-    }
-
-    curl_easy_setopt(curl_->handle, CURLOPT_MIMEPOST, curl_->multipart);
-    hasBodyOrPayload_ = true;
+    content_ = multipart;
 }
 
 void Session::SetMultipart(Multipart&& multipart) {
-    SetMultipart(multipart);
+    content_ = std::move(multipart);
 }
 
 void Session::SetRedirect(const Redirect& redirect) {
@@ -450,15 +454,11 @@ void Session::SetCookies(const Cookies& cookies) {
 }
 
 void Session::SetBody(const Body& body) {
-    hasBodyOrPayload_ = true;
-    curl_easy_setopt(curl_->handle, CURLOPT_POSTFIELDSIZE_LARGE, static_cast<curl_off_t>(body.str().length()));
-    curl_easy_setopt(curl_->handle, CURLOPT_POSTFIELDS, body.c_str());
+    content_ = body;
 }
 
 void Session::SetBody(Body&& body) {
-    hasBodyOrPayload_ = true;
-    curl_easy_setopt(curl_->handle, CURLOPT_POSTFIELDSIZE_LARGE, static_cast<curl_off_t>(body.str().length()));
-    curl_easy_setopt(curl_->handle, CURLOPT_COPYPOSTFIELDS, body.c_str());
+    content_ = std::move(body);
 }
 
 void Session::SetLowSpeed(const LowSpeed& low_speed) {
@@ -517,7 +517,7 @@ void Session::SetSslOptions(const SslOptions& options) {
 #endif
     curl_easy_setopt(curl_->handle, CURLOPT_SSL_VERIFYPEER, options.verify_peer ? ON : OFF);
     curl_easy_setopt(curl_->handle, CURLOPT_SSL_VERIFYHOST, options.verify_host ? 2L : 0L);
-#if LIBCURL_VERSION_NUM >= 0x072900
+#if LIBCURL_VERSION_NUM >= 0x072900 // 7.41.0
     curl_easy_setopt(curl_->handle, CURLOPT_SSL_VERIFYSTATUS, options.verify_status ? ON : OFF);
 #endif
 
@@ -577,11 +577,11 @@ void Session::SetInterface(const Interface& iface) {
 }
 
 void Session::SetLocalPort(const LocalPort& local_port) {
-    curl_easy_setopt(curl_->handle, CURLOPT_LOCALPORT, local_port);
+    curl_easy_setopt(curl_->handle, CURLOPT_LOCALPORT, static_cast<long>(static_cast<uint16_t>(local_port)));
 }
 
 void Session::SetLocalPortRange(const LocalPortRange& local_port_range) {
-    curl_easy_setopt(curl_->handle, CURLOPT_LOCALPORTRANGE, local_port_range);
+    curl_easy_setopt(curl_->handle, CURLOPT_LOCALPORTRANGE, static_cast<long>(static_cast<uint16_t>(local_port_range)));
 }
 
 void Session::SetHttpVersion(const HttpVersion& version) {
@@ -624,7 +624,6 @@ void Session::SetHttpVersion(const HttpVersion& version) {
 
         default: // Should not happen
             throw std::invalid_argument("Invalid/Unknown HTTP version type.");
-            break;
     }
 }
 
@@ -651,15 +650,15 @@ void Session::SetAcceptEncoding(AcceptEncoding&& accept_encoding) {
 }
 
 cpr_off_t Session::GetDownloadFileLength() {
-    cpr_off_t downloadFileLenth = -1;
+    cpr_off_t downloadFileLength = -1;
     curl_easy_setopt(curl_->handle, CURLOPT_URL, url_.c_str());
 
     const std::string protocol = url_.str().substr(0, url_.str().find(':'));
     if (proxies_.has(protocol)) {
         curl_easy_setopt(curl_->handle, CURLOPT_PROXY, proxies_[protocol].c_str());
         if (proxyAuth_.has(protocol)) {
-            curl_easy_setopt(curl_->handle, CURLOPT_PROXYAUTH, CURLAUTH_ANY);
-            curl_easy_setopt(curl_->handle, CURLOPT_PROXYUSERPWD, proxyAuth_[protocol]);
+            curl_easy_setopt(curl_->handle, CURLOPT_PROXYUSERNAME, proxyAuth_.GetUsername(protocol));
+            curl_easy_setopt(curl_->handle, CURLOPT_PROXYPASSWORD, proxyAuth_.GetPassword(protocol));
         }
     }
 
@@ -670,10 +669,10 @@ cpr_off_t Session::GetDownloadFileLength() {
         long status_code{};
         curl_easy_getinfo(curl_->handle, CURLINFO_RESPONSE_CODE, &status_code);
         if (200 == status_code) {
-            curl_easy_getinfo(curl_->handle, CURLINFO_CONTENT_LENGTH_DOWNLOAD_T, &downloadFileLenth);
+            curl_easy_getinfo(curl_->handle, CURLINFO_CONTENT_LENGTH_DOWNLOAD_T, &downloadFileLength);
         }
     }
-    return downloadFileLenth;
+    return downloadFileLength;
 }
 
 void Session::ResponseStringReserve(size_t size) {
@@ -789,7 +788,7 @@ void Session::PrepareDelete() {
 void Session::PrepareGet() {
     // In case there is a body or payload for this request, we create a custom GET-Request since a
     // GET-Request with body is based on the HTTP RFC **not** a leagal request.
-    if (hasBodyOrPayload_) {
+    if (hasBodyOrPayload()) {
         curl_easy_setopt(curl_->handle, CURLOPT_NOBODY, 0L);
         curl_easy_setopt(curl_->handle, CURLOPT_CUSTOMREQUEST, "GET");
     } else {
@@ -822,10 +821,10 @@ void Session::PreparePost() {
     curl_easy_setopt(curl_->handle, CURLOPT_NOBODY, 0L);
 
     // In case there is no body or payload set it to an empty post:
-    if (hasBodyOrPayload_) {
+    if (hasBodyOrPayload()) {
         curl_easy_setopt(curl_->handle, CURLOPT_CUSTOMREQUEST, nullptr);
     } else {
-        curl_easy_setopt(curl_->handle, CURLOPT_POSTFIELDS, readcb_.callback ? nullptr : "");
+        curl_easy_setopt(curl_->handle, CURLOPT_POSTFIELDS, cbs_->readcb_.callback ? nullptr : "");
         curl_easy_setopt(curl_->handle, CURLOPT_CUSTOMREQUEST, "POST");
     }
     prepareCommon();
@@ -833,7 +832,7 @@ void Session::PreparePost() {
 
 void Session::PreparePut() {
     curl_easy_setopt(curl_->handle, CURLOPT_NOBODY, 0L);
-    if (!hasBodyOrPayload_ && readcb_.callback) {
+    if (!hasBodyOrPayload() && cbs_->readcb_.callback) {
         /**
          * Yes, this one has to be CURLOPT_POSTFIELDS even if we are performing a PUT request.
          * In case we don't set this one, performing a POST-request with PUT won't work.
@@ -872,15 +871,12 @@ Response Session::Complete(CURLcode curl_error) {
     Cookies cookies = util::parseCookies(raw_cookies);
     curl_slist_free_all(raw_cookies);
 
-    // Reset the has no body property:
-    hasBodyOrPayload_ = false;
-
     std::string errorMsg = curl_->error.data();
     return Response(curl_, std::move(response_string_), std::move(header_string_), std::move(cookies), Error(curl_error, std::move(errorMsg)));
 }
 
 Response Session::CompleteDownload(CURLcode curl_error) {
-    if (!headercb_.callback) {
+    if (!cbs_->headercb_.callback) {
         curl_easy_setopt(curl_->handle, CURLOPT_HEADERFUNCTION, nullptr);
         curl_easy_setopt(curl_->handle, CURLOPT_HEADERDATA, 0);
     }
@@ -895,7 +891,10 @@ Response Session::CompleteDownload(CURLcode curl_error) {
 }
 
 void Session::AddInterceptor(const std::shared_ptr<Interceptor>& pinterceptor) {
-    interceptors_.push(pinterceptor);
+    // Shall only add before first interceptor run
+    assert(current_interceptor_ == interceptors_.end());
+    interceptors_.push_back(pinterceptor);
+    first_interceptor_ = interceptors_.begin();
 }
 
 Response Session::proceed() {
@@ -903,11 +902,89 @@ Response Session::proceed() {
     return makeRequest();
 }
 
-Response Session::intercept() {
-    // At least one interceptor exists -> Execute its intercept function
-    const std::shared_ptr<Interceptor> interceptor = interceptors_.front();
-    interceptors_.pop();
-    return interceptor->intercept(*this);
+const std::optional<Response> Session::intercept() {
+    if (current_interceptor_ == interceptors_.end()) {
+        current_interceptor_ = first_interceptor_;
+    } else {
+        current_interceptor_++;
+    }
+
+    if (current_interceptor_ != interceptors_.end()) {
+        auto icpt = current_interceptor_;
+        // Nested makeRequest() start at first_interceptor_, thus excluding previous interceptors.
+        first_interceptor_ = current_interceptor_;
+        ++first_interceptor_;
+
+        const std::optional<Response> r = (*current_interceptor_)->intercept(*this);
+
+        first_interceptor_ = icpt;
+
+        return r;
+    }
+    return std::nullopt;
+}
+
+void Session::prepareBodyPayloadOrMultipart() const {
+    // Either a body, multipart or a payload is allowed.
+
+    if (std::holds_alternative<cpr::Payload>(content_)) {
+        const std::string payload = std::get<cpr::Payload>(content_).GetContent(*curl_);
+        curl_easy_setopt(curl_->handle, CURLOPT_POSTFIELDSIZE_LARGE, static_cast<curl_off_t>(payload.length()));
+        curl_easy_setopt(curl_->handle, CURLOPT_COPYPOSTFIELDS, payload.c_str());
+    } else if (std::holds_alternative<cpr::Body>(content_)) {
+        const std::string& body = std::get<cpr::Body>(content_).str();
+        curl_easy_setopt(curl_->handle, CURLOPT_POSTFIELDSIZE_LARGE, static_cast<curl_off_t>(body.length()));
+        curl_easy_setopt(curl_->handle, CURLOPT_COPYPOSTFIELDS, body.c_str());
+    } else if (std::holds_alternative<cpr::Multipart>(content_)) {
+        // Make sure, we have a empty multipart to start with:
+        if (curl_->multipart) {
+            curl_mime_free(curl_->multipart);
+        }
+        curl_->multipart = curl_mime_init(curl_->handle);
+
+        // Add all multipart pieces:
+        const cpr::Multipart& multipart = std::get<cpr::Multipart>(content_);
+        for (const Part& part : multipart.parts) {
+            if (part.is_file) {
+                for (const File& file : part.files) {
+                    curl_mimepart* mimePart = curl_mime_addpart(curl_->multipart);
+                    if (!part.content_type.empty()) {
+                        curl_mime_type(mimePart, part.content_type.c_str());
+                    }
+
+                    curl_mime_filedata(mimePart, file.filepath.c_str());
+                    curl_mime_name(mimePart, part.name.c_str());
+
+                    if (file.hasOverridenFilename()) {
+                        curl_mime_filename(mimePart, file.overriden_filename.c_str());
+                    } else {
+                        // NOLINTNEXTLINE (misc-include-cleaner)
+                        curl_mime_filename(mimePart, fs::path(file.filepath).filename().string().c_str());
+                    }
+                }
+            } else {
+                curl_mimepart* mimePart = curl_mime_addpart(curl_->multipart);
+                if (!part.content_type.empty()) {
+                    curl_mime_type(mimePart, part.content_type.c_str());
+                }
+                if (part.is_buffer) {
+                    // Do not use formdata, to prevent having to use reinterpreter_cast:
+                    curl_mime_name(mimePart, part.name.c_str());
+                    curl_mime_data(mimePart, part.data, part.datalen);
+                    curl_mime_filename(mimePart, part.value.c_str());
+                } else {
+                    curl_mime_name(mimePart, part.name.c_str());
+                    curl_mime_data(mimePart, part.value.c_str(), CURL_ZERO_TERMINATED);
+                }
+            }
+        }
+
+        curl_easy_setopt(curl_->handle, CURLOPT_MIMEPOST, curl_->multipart);
+    }
+}
+
+[[nodiscard]] bool Session::hasBodyOrPayload() const {
+    return std::holds_alternative<cpr::Body>(content_) || std::holds_alternative<cpr::Payload>(content_);
 }
 
 // clang-format off
@@ -957,18 +1034,18 @@ void Session::SetOption(const Range& range) { SetRange(range); }
 void Session::SetOption(const MultiRange& multi_range) { SetMultiRange(multi_range); }
 void Session::SetOption(const ReserveSize& reserve_size) { SetReserveSize(reserve_size.size); }
 void Session::SetOption(const AcceptEncoding& accept_encoding) { SetAcceptEncoding(accept_encoding); }
-void Session::SetOption(AcceptEncoding&& accept_encoding) { SetAcceptEncoding(accept_encoding); }
+void Session::SetOption(AcceptEncoding&& accept_encoding) { SetAcceptEncoding(std::move(accept_encoding)); }
 // clang-format on
 
 void Session::SetCancellationParam(std::shared_ptr<std::atomic_bool> param) {
-    cancellationcb_ = CancellationCallback{std::move(param)};
+    cbs_->cancellationcb_ = CancellationCallback{std::move(param)};
     isCancellable = true;
-#if LIBCURL_VERSION_NUM < 0x072000
+#if LIBCURL_VERSION_NUM < 0x072000 // 7.32.0
     curl_easy_setopt(curl_->handle, CURLOPT_PROGRESSFUNCTION, cpr::util::progressUserFunction<CancellationCallback>);
-    curl_easy_setopt(curl_->handle, CURLOPT_PROGRESSDATA, &cancellationcb_);
+    curl_easy_setopt(curl_->handle, CURLOPT_PROGRESSDATA, &cbs_->cancellationcb_);
 #else
     curl_easy_setopt(curl_->handle, CURLOPT_XFERINFOFUNCTION, cpr::util::progressUserFunction<CancellationCallback>);
-    curl_easy_setopt(curl_->handle, CURLOPT_XFERINFODATA, &cancellationcb_);
+    curl_easy_setopt(curl_->handle, CURLOPT_XFERINFODATA, &cbs_->cancellationcb_);
 #endif
     curl_easy_setopt(curl_->handle, CURLOPT_NOPROGRESS, 0L);
 }
