@@ -1,4 +1,5 @@
 #include "httpServer.hpp"
+#include <array>
 #include <chrono>
 #include <string>
 #include <system_error>
@@ -142,12 +143,10 @@ void HttpServer::OnRequestLowSpeedBytes(mg_connection* conn, mg_http_message* /*
 }
 
 void HttpServer::OnRequestBasicCookies(mg_connection* conn, mg_http_message* /*msg*/) {
-    time_t expires_time = 3905119080; // Expires=Wed, 30 Sep 2093 03:18:00 GMT
-    std::array<char, EXPIRES_STRING_SIZE> expires_string;
-    std::strftime(expires_string.data(), expires_string.size(), "%a, %d %b %Y %H:%M:%S GMT", gmtime(&expires_time));
+    const std::string expires = "Wed, 30 Sep 2093 03:18:00 GMT";
 
-    std::string cookie1{"SID=31d4d96e407aad42; Expires=" + std::string(expires_string.data()) + "; Secure"};
-    std::string cookie2{"lang=en-US; Expires=" + std::string(expires_string.data()) + "; Secure"};
+    std::string cookie1{"SID=31d4d96e407aad42; Expires=" + expires + "; Secure"};
+    std::string cookie2{"lang=en-US; Expires=" + expires + "; Secure"};
     std::string headers =
             "Content-Type: text/html\r\n"
             "Set-Cookie: " +
@@ -161,12 +160,10 @@ void HttpServer::OnRequestBasicCookies(mg_connection* conn, mg_http_message* /*m
 }
 
 void HttpServer::OnRequestEmptyCookies(mg_connection* conn, mg_http_message* /*msg*/) {
-    time_t expires_time = 3905119080; // Expires=Wed, 30 Sep 2093 03:18:00 GMT
-    std::array<char, EXPIRES_STRING_SIZE> expires_string;
-    std::strftime(expires_string.data(), sizeof(expires_string), "%a, %d %b %Y %H:%M:%S GMT", gmtime(&expires_time));
+    const std::string expires = "Wed, 30 Sep 2093 03:18:00 GMT";
 
-    std::string cookie1{"SID=; Expires=" + std::string(expires_string.data()) + "; Secure"};
-    std::string cookie2{"lang=; Expires=" + std::string(expires_string.data()) + "; Secure"};
+    std::string cookie1{"SID=; Expires=" + expires + "; Secure"};
+    std::string cookie2{"lang=; Expires=" + expires + "; Secure"};
     std::string headers =
             "Content-Type: text/html\r\n"
             "Set-Cookie: " +
@@ -180,7 +177,7 @@ void HttpServer::OnRequestEmptyCookies(mg_connection* conn, mg_http_message* /*m
 }
 
 void HttpServer::OnRequestCookiesReflect(mg_connection* conn, mg_http_message* msg) {
-    mg_str* request_cookies;
+    mg_str* request_cookies{nullptr};
     if ((request_cookies = mg_http_get_header(msg, "Cookie")) == nullptr) {
         std::string errorMessage{"Cookie not found"};
         SendError(conn, 400, errorMessage);
@@ -192,19 +189,17 @@ void HttpServer::OnRequestCookiesReflect(mg_connection* conn, mg_http_message* m
 }
 
 void HttpServer::OnRequestRedirectionWithChangingCookies(mg_connection* conn, mg_http_message* msg) {
-    time_t expires_time = 3905119080; // Expires=Wed, 30 Sep 2093 03:18:00 GMT
-    std::array<char, EXPIRES_STRING_SIZE> expires_string;
-    std::strftime(expires_string.data(), sizeof(expires_string), "%a, %d %b %Y %H:%M:%S GMT", gmtime(&expires_time));
+    const std::string expires = "Wed, 30 Sep 2093 03:18:00 GMT";
 
-    mg_str* request_cookies;
+    mg_str* request_cookies{nullptr};
     std::string cookie_str;
     if ((request_cookies = mg_http_get_header(msg, "Cookie")) != nullptr) {
         cookie_str = std::string{request_cookies->ptr, request_cookies->len};
     }
 
     if (cookie_str.find("SID=31d4d96e407aad42") == std::string::npos) {
-        std::string cookie1{"SID=31d4d96e407aad42; Expires=" + std::string(expires_string.data()) + "; Secure"};
-        std::string cookie2{"lang=en-US; Expires=" + std::string(expires_string.data()) + "; Secure"};
+        std::string cookie1{"SID=31d4d96e407aad42; Expires=" + expires + "; Secure"};
+        std::string cookie2{"lang=en-US; Expires=" + expires + "; Secure"};
         std::string headers =
                 "Content-Type: text/html\r\n"
                 "Location: http://127.0.0.1:61936/redirection_with_changing_cookies.html\r\n"
@@ -394,12 +389,12 @@ void HttpServer::OnRequestUrlPost(mg_connection* conn, mg_http_message* msg) {
 
     std::string headers = "Content-Type: application/json\r\n";
 
-    char x[100];
-    char y[100];
-    mg_http_get_var(&(msg->body), "x", x, sizeof(x));
-    mg_http_get_var(&(msg->body), "y", y, sizeof(y));
-    std::string x_string{x};
-    std::string y_string{y};
+    std::array<char, 100> x{0};
+    std::array<char, 100> y{0};
+    mg_http_get_var(&(msg->body), "x", x.data(), x.size());
+    mg_http_get_var(&(msg->body), "y", y.data(), y.size());
+    std::string x_string{x.data()};
+    std::string y_string{y.data()};
     std::string response;
     if (y_string.empty()) {
         response = std::string{
@@ -418,7 +413,7 @@ void HttpServer::OnRequestUrlPost(mg_connection* conn, mg_http_message* msg) {
                 y_string +
                 ",\n"
                 "  \"sum\": " +
-                std::to_string(atoi(x) + atoi(y)) +
+                std::to_string(atoi(x.data()) + atoi(y.data())) +
                 "\n"
                 "}"};
     }
@@ -475,6 +470,26 @@ void HttpServer::OnRequestFormPost(mg_connection* conn, mg_http_message* msg) {
     mg_http_reply(conn, 201, headers.c_str(), response.c_str());
 }
 
+void HttpServer::OnRequestFileUploadPost(mg_connection* conn, mg_http_message* msg) {
+    size_t pos{0};
+    mg_http_part part{};
+
+    std::string headers = "Content-Type: application/json\r\n";
+    std::string response{};
+    response += "{\n";
+    while ((pos = mg_http_next_multipart(msg->body, pos, &part)) > 0) {
+        response += "  \"" + std::string(part.name.ptr, part.name.len) + "\": \"";
+        if (!std::string(part.filename.ptr, part.filename.len).empty()) {
+            response += std::string(part.filename.ptr, part.filename.len) + "=";
+        }
+        response += std::string(part.body.ptr, part.body.len) + "\",\n";
+    }
+    response.erase(response.find_last_not_of(",\n") + 1);
+    response += "\n}";
+
+    mg_http_reply(conn, 201, headers.c_str(), response.c_str());
+}
+
 void HttpServer::OnRequestDelete(mg_connection* conn, mg_http_message* msg) {
     bool has_json_header = false;
     for (mg_http_header& header : msg->headers) {
@@ -519,12 +534,12 @@ void HttpServer::OnRequestDeleteNotAllowed(mg_connection* conn, mg_http_message*
 
 void HttpServer::OnRequestPut(mg_connection* conn, mg_http_message* msg) {
     if (std::string{msg->method.ptr, msg->method.len} == std::string{"PUT"}) {
-        char x[100];
-        char y[100];
-        mg_http_get_var(&(msg->body), "x", x, sizeof(x));
-        mg_http_get_var(&(msg->body), "y", y, sizeof(y));
-        std::string x_string = std::string{x};
-        std::string y_string = std::string{y};
+        std::array<char, 100> x{0};
+        std::array<char, 100> y{0};
+        mg_http_get_var(&(msg->body), "x", x.data(), x.size());
+        mg_http_get_var(&(msg->body), "y", y.data(), y.size());
+        std::string x_string{x.data()};
+        std::string y_string{y.data()};
         std::string headers = "Content-Type: application/json\r\n";
         std::string response;
         if (y_string.empty()) {
@@ -544,7 +559,7 @@ void HttpServer::OnRequestPut(mg_connection* conn, mg_http_message* msg) {
                     y_string +
                     ",\n"
                     "  \"sum\": " +
-                    std::to_string(atoi(x) + atoi(y)) +
+                    std::to_string(atoi(x.data()) + atoi(y.data())) +
                     "\n"
                     "}"};
         }
@@ -591,12 +606,12 @@ void HttpServer::OnRequestPutNotAllowed(mg_connection* conn, mg_http_message* ms
 
 void HttpServer::OnRequestPatch(mg_connection* conn, mg_http_message* msg) {
     if (std::string{msg->method.ptr, msg->method.len} == std::string{"PATCH"}) {
-        char x[100];
-        char y[100];
-        mg_http_get_var(&(msg->body), "x", x, sizeof(x));
-        mg_http_get_var(&(msg->body), "y", y, sizeof(y));
-        std::string x_string = std::string{x};
-        std::string y_string = std::string{y};
+        std::array<char, 100> x{0};
+        std::array<char, 100> y{0};
+        mg_http_get_var(&(msg->body), "x", x.data(), x.size());
+        mg_http_get_var(&(msg->body), "y", y.data(), y.size());
+        std::string x_string{x.data()};
+        std::string y_string{y.data()};
         std::string headers = "Content-Type: application/json\r\n";
         std::string response;
         if (y_string.empty()) {
@@ -616,7 +631,7 @@ void HttpServer::OnRequestPatch(mg_connection* conn, mg_http_message* msg) {
                     y_string +
                     ",\n"
                     "  \"sum\": " +
-                    std::to_string(atoi(x) + atoi(y)) +
+                    std::to_string(atoi(x.data()) + atoi(y.data())) +
                     "\n"
                     "}"};
         }
@@ -887,6 +902,8 @@ void HttpServer::OnRequest(mg_connection* conn, mg_http_message* msg) {
         OnRequestJsonPost(conn, msg);
     } else if (uri == "/form_post.html") {
         OnRequestFormPost(conn, msg);
+    } else if (uri == "/post_file_upload.html") {
+        OnRequestFileUploadPost(conn, msg);
     } else if (uri == "/post_reflect.html") {
         OnRequestPostReflect(conn, msg);
     } else if (uri == "/delete.html") {
